#ifndef HARKTOOL4_UTILS_STD_HPP
#define HARKTOOL4_UTILS_STD_HPP

#include <istream>
#include <fstream>
#include <vector>
#include <type_traits>
#include <limits>
#include <algorithm>
#include <iterator>

namespace harktool4
{

extern void* enabler; // for SFINAE

/** @brief 指定されたファイルのサイズを計算して返す.
 * この関数の呼び出し前後でifsの読み込み位置は変化しない。
 */
std::streampos GetFileSize(std::ifstream& ifs);

/** @brief 指定された入力ストリームから指定された個数だけfloatを読み込んで返す.
 */
std::vector<float> ReadFloatArray(std::istream& is, size_t n);

/** @brief 指定された入力ストリームから可能な限りfloatを読み込んで返す.
 */
std::vector<float> ReadFloatArray(std::istream& is);

/** @brief コンテナの要素をstrで連結する.
 * 例えば{1, 2, 3}というコンテナと", "という文字列が渡されたら
 * "1, 2, 3"という文字列を生成して返す。
 * @param first 連結する先頭の要素へのイテレータ
 * @param last 連結する最後の要素を1つ超えた要素へのイテレータ
 * @param str 連結する文字列
 */
template <typename InputIterator, typename CharT>
std::basic_string<CharT> Join(InputIterator first, InputIterator last, const CharT* str);

template <typename InputIterator, typename CharT>
std::basic_string<CharT> Join(InputIterator first, InputIterator last, const CharT* str)
{
    std::basic_ostringstream<CharT> oss;
    InputIterator it = first;

    if (it != last)
    {
        oss << *it;
        ++it;
    }
    for (; it != last; ++it)
    {
        oss << str << *it;
    }

    return oss.str();
}

/* 多次元ベクトルの型を表す */
template <template <typename T, typename Allocator=std::allocator<T>> class V, typename T, size_t D>
struct MultiDimVector
{
    typedef typename MultiDimVector<V, T, D - 1>::type value_type;
    typedef V<value_type, std::allocator<value_type> > type;
};

template <template <typename T, typename Allocator=std::allocator<T>> class V, typename T>
struct MultiDimVector<V, T, 0>
{
    typedef T type;
};

/** 指定した要素を持つ多次元ベクトルを作る.
 *
 * example: to make float[2][3][4]:
 *   auto vect = CreateMultiDimVector<std::vector, float>(2, 3, 4);
 */
template <template <typename T, typename Allocator=std::allocator<T>> class V, typename T, typename Dim,
         typename std::enable_if<std::is_integral<Dim>::value>::type*& = enabler>
typename MultiDimVector<V, T, 1>::type
CreateMultiDimVector(const T& initial_value, Dim dim)
{
    return typename MultiDimVector<V, T, 1>::type(dim, initial_value);
    /*
     * typename std::enable_if<std::is_integral<Dim>::value>::type* = 0
     * は次元指定に整数のみを許すための記述。Dimが整数型以外の場合はこの関数は定義されない。
     */
}

template <template <typename T, typename Allocator=std::allocator<T>> class V, typename T, typename Dim, typename ... Dims,
         typename std::enable_if<std::is_integral<Dim>::value>::type*& = enabler>
typename MultiDimVector<V, T, 1 + sizeof...(Dims)>::type
CreateMultiDimVector(const T& initial_value, Dim dim, Dims ... dims)
{
    return typename MultiDimVector<V, T, 1 + sizeof...(Dims)>::type(
            dim, CreateMultiDimVector<V, T, Dims ...>(initial_value, dims ...));
}

/** 与えられた型が加算可能かどうかを調べる.
 * example:
 *   if (AreAddable<T, U>::value) { // T can be added with U. }
 */
template <typename Lhs, typename Rhs>
struct AreAddable
{
private:
    template <typename T, typename U>
    static auto Check(int) -> decltype(*reinterpret_cast<T*>(0) + *reinterpret_cast<U*>(0), std::true_type());
    template <typename T, typename U>
    static std::false_type Check(...);

public:
    static const bool value = std::is_same<decltype(Check<Lhs, Rhs>(0)), std::true_type>::value;
};

template <typename T>
struct IsAddable
{
public:
    static const bool value = AreAddable<T, T>::value;
};

/** 与えられた型が減算可能かどうかを調べる.
 * example:
 *   if (AreAddable<T, U>::value) { // T can be sbtracted with U. }
 */
template <typename Lhs, typename Rhs>
struct AreSubtractable
{
private:
    template <typename T, typename U>
    static auto Check(int) -> decltype(*reinterpret_cast<T*>(0) - *reinterpret_cast<U*>(0), std::true_type());
    template <typename T, typename U>
    static std::false_type Check(...);

public:
    static const bool value = std::is_same<decltype(Check<Lhs, Rhs>(0)), std::true_type>::value;
};

template <typename T>
struct IsSubtractable
{
public:
    static const bool value = AreSubtractable<T, T>::value;
};

/** 与えられた型が乗算可能かどうかを調べる.
 * example:
 *   if (AreAddable<T, U>::value) { // T can be multiplied with U. }
 */
template <typename Lhs, typename Rhs>
struct AreMultipliable
{
private:
    template <typename T, typename U>
    static auto Check(int) -> decltype(*reinterpret_cast<T*>(0) * *reinterpret_cast<U*>(0), std::true_type());
    template <typename T, typename U>
    static std::false_type Check(...);

public:
    static const bool value = std::is_same<decltype(Check<Lhs, Rhs>(0)), std::true_type>::value;
};

template <typename T>
struct IsMultipliable
{
public:
    static const bool value = AreMultipliable<T, T>::value;
};

/** 与えられた型が除算可能かどうかを調べる.
 * example:
 *   if (AreAddable<T, U>::value) { // T can be devided with U. }
 */
template <typename Lhs, typename Rhs>
struct AreDividable
{
private:
    template <typename T, typename U>
    static auto Check(int) -> decltype(*reinterpret_cast<T*>(0) / *reinterpret_cast<U*>(0), std::true_type());
    template <typename T, typename U>
    static std::false_type Check(...);

public:
    static const bool value = std::is_same<decltype(Check<Lhs, Rhs>(0)), std::true_type>::value;
};

template <typename T>
struct IsDividable
{
public:
    static const bool value = AreDividable<T, T>::value;
};

/** 与えられたコンテナのvalue_typeをたどり、最初に見つかったPredを満たすvalue_typeを返す.
 * example:
 *   // 次元数がわからない多次元配列の要素型を取得
 *   typedef vector<vector<double>> UnknownDimensionVector;
 *   FirstValueTypeIf<UnknownDimensionVector, IsAddable>::type v; // v is double
 *
 *   // When Pred is a predicate such as Pred<SomeType>::value is false, following statement is ill-formed
 *   FirstValueTypeIf<vector<SomeType>, Pred>::type v; // error!
 */
template <typename T, template <typename> class Pred, bool = Pred<T>::value>
struct FirstValueTypeIf
{
    typedef T type;
};

template <typename T, template <typename> class Pred>
struct FirstValueTypeIf<T, Pred, false>
{
    typedef typename FirstValueTypeIf<typename T::value_type, Pred>::type type;
};

/** @brief 多次元配列の要素の総和を求める. */
template <typename T, typename std::enable_if<IsAddable<T>::value>::type*& = enabler>
T
SumAll(const T& value)
{
    return value;
}

template <typename T, typename std::enable_if<!IsAddable<T>::value>::type*& = enabler>
typename FirstValueTypeIf<T, IsAddable>::type
SumAll(const T& v)
{
    typename FirstValueTypeIf<T, IsAddable>::type sum = 0;
    for (auto it = v.begin(); it != v.end(); ++it)
    {
        sum += SumAll(*it);
    }
    return sum;
}

/** @brief 多次元配列のすべての要素を指定した数値で割る. */
template <typename T, typename U, typename std::enable_if<AreDividable<T, U>::value>::type*& = enabler>
void DivideAll(T& lhs, const U& rhs)
{
    lhs /= rhs;
}

template <typename T, typename U, typename std::enable_if<!AreDividable<T, U>::value>::type*& = enabler>
void DivideAll(T& v, const U& rhs)
{
    for (auto it = v.begin(); it != v.end(); ++it)
    {
        DivideAll(*it, rhs);
    }
}

/** @brief 多次元配列のすべての要素に指定した数値を掛ける. */
template <typename T, typename U, typename std::enable_if<AreMultipliable<T, U>::value>::type*& = enabler>
void MultiplyAll(T& lhs, const U& rhs)
{
    lhs *= rhs;
}

template <typename T, typename U, typename std::enable_if<!AreMultipliable<T, U>::value>::type*& = enabler>
void MultiplyAll(T& v, const U& rhs)
{
    for (auto it = v.begin(); it != v.end(); ++it)
    {
        MultiplyAll(*it, rhs);
    }
}

/** @brief 指定された要素列で最大の要素へのイテレータを返す.
 * begin_とend_で指定された要素列から、最大の要素を検索し、
 * その要素へのイテレータを返す。
 */
template <typename InputIterator>
InputIterator FindTop(InputIterator begin_, InputIterator end_)
{
    auto it = begin_;
    auto max_index = it;
    ++it;

    for (; it != end_; ++it)
    {
        if (*max_index < *it)
        {
            max_index = it;
        }
    }

    return max_index;
}

/** @brief 指定された要素列から上位n個の要素を検索する.
 * begin_とend_で指定された要素列から、大きい順にn個の要素を検索し、
 * out_begin_で指定された場所に発見した要素へのイテレータを格納する。
 */
template <typename InputIterator, typename OutputIterator>
void FindTopN(InputIterator begin_, InputIterator end_, size_t n, OutputIterator out_begin_)
{
    // 調査対象のイテレータ集合
    std::vector<InputIterator> remain_elements(std::distance(begin_, end_));
    // 入力イテレータ数
    const auto input_size = remain_elements.size();

    // 調査対象のイテレータを、remain_elementsへコピー
    {
        auto it = begin_;
        for (size_t i = 0; i < input_size; ++i)
        {
            remain_elements[i] = it;
            ++it;
        }
    }

    // remain_elementsから上位n個の要素の位置を取得し、out_begin_へコピー
    for (size_t i = 0; i < n && i < input_size; ++i)
    {
        assert(!remain_elements.empty());

        auto it = remain_elements.begin();
        auto max_index = it;
        ++it;

        for (; it != remain_elements.end(); ++it)
        {
            if (**max_index < **it)
            {
                max_index = it;
            }
        }

        *out_begin_ = *max_index;
        ++out_begin_;
        remain_elements.erase(max_index);
    }
}

template <typename T>
struct HasNumericLimitsLowest
{
    template <typename U>
    static auto Check(int) -> decltype(reinterpret_cast<std::numeric_limits<U>*>(0)->lowest(), std::true_type());
    template <typename U>
    static std::false_type Check(...);

    static const bool value = std::is_same<decltype(Check<T>(0)), std::true_type>::value;
};

/** std::numeric_limits<T>::lowest()相当の機能を提供する.
 * std::numeric_limits<T>::lowest()が提供されていない古い処理系に対応するための関数。
 * std::numeric_limits<T>::lowest()が提供されているなら、それを用いる。
 *
 * GCC 4.5以前ではconstexprが使えないため、コメントアウトしている。
 * constexprを有効にしたほうが効率がいいプログラムになる。
 */
template <typename T, typename std::enable_if<HasNumericLimitsLowest<T>::value>::type*& = enabler>
/*constexpr */T NumericLimitsLowest()
{
    return std::numeric_limits<T>::lowest();
}

template <typename T, typename std::enable_if<!HasNumericLimitsLowest<T>::value && std::is_floating_point<T>::value>::type*& = enabler>
/*constexpr */T NumericLimitsLowest()
{
    return -std::numeric_limits<T>::max();
}

template <typename T, typename std::enable_if<!HasNumericLimitsLowest<T>::value && std::is_integral<T>::value>::type*& = enabler>
/*constexpr */T NumericLimitsLowest()
{
    return std::numeric_limits<T>::min();
}

}

#endif // HARKTOOL4_UTILS_STD_HPP
